@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2017 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2018 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -79,12 +79,12 @@ int wrapperGetUNCFilePath(const TCHAR *path, int advice) {
         result = GetDriveType(drive);
         if (result == DRIVE_REMOTE) {
             if (advice == 0) {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ADVICE, TEXT("The following path in your Wrapper configuration file is to a mapped Network\n  Drive.  Using mapped network drives is not recommeded as they will fail to\n  be resolved correctly under certain circumstances.  Please consider using\n  UNC paths (\\\\<host>\\<share>\\path). Additional refrences will be ignored.\n  Path: %s"), path);
+                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ADVICE, TEXT("The following path in your Wrapper configuration file is to a mapped Network\n  Drive.  Using mapped network drives is not recommended as they will fail to\n  be resolved correctly under certain circumstances.  Please consider using\n  UNC paths (\\\\<host>\\<share>\\path). Additional references will be ignored.\n  Path: %s"), path);
                 advice++;
             }
         } else if (result == DRIVE_NO_ROOT_DIR) {
             if (advice == 0) {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ADVICE, TEXT("The following path in your Wrapper configuration file could not be resolved.\n  Please make sure the path exists.  If the path is a network share, it may be\n  that the current user is unable to resolve it.  Please consider using UNC\n  paths (\\\\<host>\\<share>\\path) or run the service as another user\n  (see wrapper.ntservice.account). Additional refrences will be ignored.\n  Path: %s"), path);
+                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ADVICE, TEXT("The following path in your Wrapper configuration file could not be resolved.\n  Please make sure the path exists.  If the path is a network share, it may be\n  that the current user is unable to resolve it.  Please consider using UNC\n  paths (\\\\<host>\\<share>\\path) or run the service as another user\n  (see wrapper.ntservice.account). Additional references will be ignored.\n  Path: %s"), path);
                 advice++;
             }
         }
@@ -93,6 +93,36 @@ int wrapperGetUNCFilePath(const TCHAR *path, int advice) {
 }
 #endif
 
+const TCHAR* getFileName(const TCHAR* path) {
+    const TCHAR* ptr1;
+#ifdef WIN32
+    const TCHAR* ptr2;
+#endif
+    
+    if (path) {
+        ptr1 = _tcsrchr(path, TEXT('/'));
+#ifdef WIN32
+        ptr2 = _tcsrchr(path, TEXT('\\'));
+        if (!ptr1 && !ptr2) {
+            return path;
+        } else if (!ptr1) {
+            return ptr2 + 1;
+        } else if (!ptr2) {
+            return ptr1 + 1;
+        } else if (&ptr2 > &ptr1) {
+            return ptr2 + 1;
+#else
+        if (!ptr1) {
+            return path;
+#endif
+        } else  {
+            return ptr1 + 1;
+        }
+    }
+    return NULL;
+}
+
+
 /**
  * Read configuration file.
  */
@@ -102,12 +132,11 @@ int configFileReader_Read(ConfigFileReader *reader,
                           int depth,
                           const TCHAR *parentFilename,
                           int parentLineNumber,
-                          const TCHAR *argCommand,
                           const TCHAR *originalWorkingDir,
                           PHashMap warnedVarMap,
+                          PHashMap ignoreVarMap,
                           int logWarnings,
-                          int logWarningLogLevel,
-                          int isDebugging)
+                          int logWarningLogLevel)
 {
     FILE *stream;
     char bufferMB[MAX_PROPERTY_NAME_VALUE_LENGTH];
@@ -129,6 +158,9 @@ int configFileReader_Read(ConfigFileReader *reader,
 #else
     char* encoding;
     char* interumEncoding;
+    TCHAR* encodingW = NULL;
+    TCHAR* interumEncodingW = NULL;
+    size_t size;
 #endif
     int includeRequired;
     int readResult = CONFIG_FILE_READER_SUCCESS;
@@ -161,7 +193,7 @@ int configFileReader_Read(ConfigFileReader *reader,
             log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Configuration file not found: %s"), filename);
 #endif
         }
-        return CONFIG_FILE_READER_FAIL;
+        return CONFIG_FILE_READER_OPEN_FAIL;
     }
 
     if (reader->debugIncludes) {
@@ -253,8 +285,8 @@ int configFileReader_Read(ConfigFileReader *reader,
     }
 
     if (depth == 0) {
-        /* At least log with LEVEL_DEBUG to help support. */
-        reader->logLevelOnOverwrite = LEVEL_DEBUG;
+        /* Default to AUTO (= -1). */
+        reader->logLevelOnOverwrite = -1;
     }
     
     /* Read all of the configurations */
@@ -278,20 +310,48 @@ int configFileReader_Read(ConfigFileReader *reader,
                     /* bufferW contains an error message. */
                     if (!reader->preload) {
                         if (depth > 0) {
-                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
+                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
                                 TEXT("%sIncluded configuration file, %s, contains a problem on line #%d and could not be read. (%s)"),
                                 (reader->debugIncludes ? TEXT("  ") : TEXT("")), filename, lineNumber, bufferW);
                         } else {
-                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
+                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ERROR,
                                 TEXT("Configuration file, %s, contains a problem on line #%d and could not be read. (%s)"), filename, lineNumber, bufferW);
                         }
+#ifndef WIN32
+                        /* On Windows there is no such problem because wide chars are always UTF-16 which should support most characters. */
+                        size = mbstowcs(NULL, encoding, 0);
+                        if (size > (size_t)0) {
+                            encodingW = malloc(sizeof(TCHAR) * size + 1);
+                            if (encodingW) {
+                                mbstowcs(encodingW, encoding, size + 1);
+                            }
+                        }
+                        size = mbstowcs(NULL, interumEncoding, 0);
+                        if (size > (size_t)0) {
+                            interumEncodingW = malloc(sizeof(TCHAR) * (size + 1));
+                            if (interumEncodingW) {
+                                mbstowcs(interumEncodingW, interumEncoding, size + 1);
+                            }
+                        }
+                        if (!compareEncodingsSysMode(encodingW, interumEncodingW)) {
+                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ADVICE,
+                                TEXT("One possible cause of failure is when the encoding of the locale used by the\n  Wrapper doesn't support certain characters of the configuration file.\n  The encoding of the current locale is '%s'. It is advised to set it\n  identical to the encoding of the configuration file (%s),\n  either by changing the default system locale or by using the wrapper.lang and\n  wrapper.lang.<platform>.encoding properties."), interumEncodingW, encodingW);
+                        }
+                        free(encodingW);
+                        free(interumEncodingW);
+#endif
                     }
                     free(bufferW);
                 } else {
                     outOfMemory(TEXT("RCF"), 1);
                 }
+                if (reader->preload) {
+                    /* On preload, ignore the line and continue. We want to load as much as possible and hopefully load the appropriate locale to read the file on the second load. */
+                    lineNumber++;
+                    continue;
+                }
                 fclose(stream);
-                return CONFIG_FILE_READER_FAIL;
+                return CONFIG_FILE_READER_HARD_FAIL;
             }
             
 #ifdef _DEBUG
@@ -317,12 +377,12 @@ int configFileReader_Read(ConfigFileReader *reader,
             }
             /* Strip any whitespace from the front of the line. */
             trimmedBuffer = bufferW;
-            while ((trimmedBuffer[0] == TEXT(' ')) || (trimmedBuffer[0] == 0x08)) {
+            while ((trimmedBuffer[0] == TEXT(' ')) || (trimmedBuffer[0] == TCHAR_TAB)) {
                 trimmedBuffer++;
             }
 
             /* If the line does not start with a comment, make sure that
-             *  any comment at the end of line are stripped.  If any any point, a
+             *  any comment at the end of line are stripped.  If at any point, a
              *  double hash, '##', is encountered it should be interpreted as a
              *  hash in the actual property rather than the beginning of a comment. */
             if (trimmedBuffer[0] != TEXT('#')) {
@@ -355,9 +415,7 @@ int configFileReader_Read(ConfigFileReader *reader,
 
             /* Strip any whitespace from the end of the line. */
             trimmedBufferLen = _tcslen(trimmedBuffer);
-            while ((trimmedBufferLen > 0) && ((trimmedBuffer[trimmedBufferLen - 1] == TEXT(' '))
-            || (trimmedBuffer[trimmedBufferLen - 1] == 0x08))) {
-
+            while ((trimmedBufferLen > 0) && ((trimmedBuffer[trimmedBufferLen - 1] == TEXT(' ')) || (trimmedBuffer[trimmedBufferLen - 1] == TCHAR_TAB))) {
                 trimmedBuffer[--trimmedBufferLen] = TEXT('\0');
             }
 
@@ -367,7 +425,7 @@ int configFileReader_Read(ConfigFileReader *reader,
                 if (trimmedBuffer[0] != TEXT('#')) {
                     /*_tprintf(TEXT("%s\n"), trimmedBuffer);*/
 
-                    if (!(*reader->callback)(reader->callbackParam, filename, lineNumber, trimmedBuffer, reader->exitOnOverwrite, reader->logLevelOnOverwrite)) {
+                    if (!(*reader->callback)(reader->callbackParam, filename, lineNumber, depth, trimmedBuffer, reader->exitOnOverwrite, reader->logLevelOnOverwrite)) {
                         readResult = CONFIG_FILE_READER_HARD_FAIL;
                         break;
                     }
@@ -405,7 +463,7 @@ int configFileReader_Read(ConfigFileReader *reader,
                             log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
                                 TEXT("Found #include file in %s: %s"), filename, c);
                         }
-                        evaluateEnvironmentVariables(c, expBuffer, MAX_PROPERTY_NAME_VALUE_LENGTH, logWarnings, warnedVarMap, logWarningLogLevel);
+                        evaluateEnvironmentVariables(c, expBuffer, MAX_PROPERTY_NAME_VALUE_LENGTH, logWarnings, warnedVarMap, logWarningLogLevel, ignoreVarMap);
                         
                         if (reader->debugIncludes && (_tcscmp(c, expBuffer) != 0)) {
                             /* Only show this log if there were any environment variables. */
@@ -427,7 +485,7 @@ int configFileReader_Read(ConfigFileReader *reader,
                         } else {
                             absoluteBuffer = malloc(sizeof(TCHAR) * size);
                             if (!absoluteBuffer) {
-                                outOfMemory(TEXT("RCF"), 1);
+                                outOfMemory(TEXT("RCF"), 2);
                             } else {
                                 if (!GetFullPathName(expBuffer, size, absoluteBuffer, NULL)) {
                                     if (reader->debugIncludes || includeRequired) {
@@ -443,7 +501,7 @@ int configFileReader_Read(ConfigFileReader *reader,
 #else
                         absoluteBuffer = malloc(sizeof(TCHAR) * (PATH_MAX + 1));
                         if (!absoluteBuffer) {
-                            outOfMemory(TEXT("RCF"), 2);
+                            outOfMemory(TEXT("RCF"), 3);
                         } else {
                             if (_trealpathN(expBuffer, absoluteBuffer, PATH_MAX + 1) == NULL) {
                                 if (reader->debugIncludes || includeRequired) {
@@ -458,10 +516,10 @@ int configFileReader_Read(ConfigFileReader *reader,
 #endif
                         if (absoluteBuffer) {
                             if (depth < MAX_INCLUDE_DEPTH) {
-                                readResult = configFileReader_Read(reader, absoluteBuffer, includeRequired, depth + 1, filename, lineNumber, argCommand, originalWorkingDir, warnedVarMap, logWarnings, logWarningLogLevel, isDebugging);
+                                readResult = configFileReader_Read(reader, absoluteBuffer, includeRequired, depth + 1, filename, lineNumber, originalWorkingDir, warnedVarMap, ignoreVarMap, logWarnings, logWarningLogLevel);
                                 if (readResult == CONFIG_FILE_READER_SUCCESS) {
                                     /* Ok continue. */
-                                } else if ((readResult == CONFIG_FILE_READER_FAIL) || (readResult == CONFIG_FILE_READER_HARD_FAIL)) {
+                                } else if ((readResult == CONFIG_FILE_READER_FAIL) || (readResult == CONFIG_FILE_READER_HARD_FAIL) || (readResult == CONFIG_FILE_READER_OPEN_FAIL)) {
                                     /* Failed. */
                                     if (includeRequired) {
                                         /* Include file was required, but we failed to read it. */
@@ -517,16 +575,20 @@ int configFileReader_Read(ConfigFileReader *reader,
                             }
                         } else if (_tcsstr(trimmedBuffer, TEXT("#properties.on_overwrite.loglevel=")) == trimmedBuffer) {
                             trimmedBuffer += 34;
-                            logLevelOnOverwrite = getLogLevelForName(trimmedBuffer);
-                            if (logLevelOnOverwrite >= LEVEL_NONE ) {
-                                /* At least log with LEVEL_DEBUG to help support. */
-                                reader->logLevelOnOverwrite = LEVEL_DEBUG;
-                            } else if (logLevelOnOverwrite != LEVEL_UNKNOWN) {
-                                reader->logLevelOnOverwrite = logLevelOnOverwrite;
-                            } else if (!reader->preload) {
-                                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
-                                    TEXT("Encountered an invalid value for directive #properties.on_overwrite.loglevel=%s (line %d).  Ignoring this directive."),
-                                    trimmedBuffer, lineNumber);
+                            if (_tcsicmp(trimmedBuffer, TEXT("AUTO")) == 0) {
+                                reader->logLevelOnOverwrite = -1;
+                            } else {
+                                logLevelOnOverwrite = getLogLevelForName(trimmedBuffer);
+                                if (logLevelOnOverwrite >= LEVEL_NONE ) {
+                                    /* At least log with LEVEL_DEBUG to help support. */
+                                    reader->logLevelOnOverwrite = LEVEL_DEBUG;
+                                } else if (logLevelOnOverwrite != LEVEL_UNKNOWN) {
+                                    reader->logLevelOnOverwrite = logLevelOnOverwrite;
+                                } else if (!reader->preload) {
+                                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
+                                        TEXT("Encountered an invalid value for directive #properties.on_overwrite.loglevel=%s (line %d).  Ignoring this directive."),
+                                        trimmedBuffer, lineNumber);
+                                }
                             }
                         } else if (strcmpIgnoreCase(trimmedBuffer, TEXT("#properties.debug")) == 0) {
                             reader->logLevelOnOverwrite = LEVEL_STATUS;
@@ -548,7 +610,7 @@ int configFileReader_Read(ConfigFileReader *reader,
      *  (we want to keep these values after preload for logging potential problems on properties defined in the command line)
      *  This is needed if directives are set at the end of the file with no properties after. */
     if (reader->preload) {
-        if (!(*reader->callback)(reader->callbackParam, NULL, -1, NULL, reader->exitOnOverwrite, reader->logLevelOnOverwrite)) {
+        if (!(*reader->callback)(reader->callbackParam, NULL, -1, depth, NULL, reader->exitOnOverwrite, reader->logLevelOnOverwrite)) {
             readResult = CONFIG_FILE_READER_HARD_FAIL;
         }
     }
@@ -571,14 +633,14 @@ int configFileReader_Read(ConfigFileReader *reader,
  * @param enableIncludes If TRUE then includes will be supported.
  * @param preload TRUE if this is being called in the preload step meaning that all errors
  *                should be suppressed.
- * @param argCommand Argument passed to the binary.
  * @param originalWorkingDir Working directory of the binary at the moment it was launched.
  * @param warnedVarMap Map of undefined environment variables for which the user was warned.
+ * @param ignoreVarMap Map of environment variables that should not be expanded.
  * @param logWarnings Flag that controls whether or not warnings will be logged.
  * @param logWarningLogLevel Log level at which any log warnings will be logged.
- * @param isDebugging Flag that controls whether or not debug output will be logged.
  *
  * @return CONFIG_FILE_READER_SUCCESS if the file was read successfully,
+ *         CONFIG_FILE_READER_OPEN_FAIL if the file could not be found or opened.
  *         CONFIG_FILE_READER_FAIL if there were any problems at all, or
  *         CONFIG_FILE_READER_HARD_FAIL if the problem should cascaded all the way up.
  */
@@ -588,12 +650,11 @@ int configFileReader(const TCHAR *filename,
                      void *callbackParam,
                      int enableIncludes,
                      int preload,
-                     const TCHAR *argCommand,
                      const TCHAR *originalWorkingDir,
                      PHashMap warnedVarMap,
+                     PHashMap ignoreVarMap,
                      int logWarnings,
-                     int logWarningLogLevel,
-                     int isDebugging) {
+                     int logWarningLogLevel) {
     ConfigFileReader reader;
     
     /* Initialize the reader. */
@@ -605,7 +666,5 @@ int configFileReader(const TCHAR *filename,
     reader.exitOnOverwrite = FALSE;
     reader.logLevelOnOverwrite = LEVEL_NONE; /* on preload, don't log anything. */
     
-    return configFileReader_Read(&reader, filename, fileRequired, 0, NULL, 0, argCommand, originalWorkingDir, warnedVarMap, logWarnings, logWarningLogLevel, isDebugging);
+    return configFileReader_Read(&reader, filename, fileRequired, 0, NULL, 0, originalWorkingDir, warnedVarMap, ignoreVarMap, logWarnings, logWarningLogLevel);
 }
-
-
