@@ -1,7 +1,7 @@
 package org.tanukisoftware.wrapper;
 
 /*
- * Copyright (c) 1999, 2017 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2018 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -13,6 +13,7 @@ package org.tanukisoftware.wrapper;
 
 import java.io.IOException;
 import java.io.InterruptedIOException;
+import java.io.UnsupportedEncodingException;
 import java.net.InetAddress;
 import java.net.Socket;
 import java.net.SocketException;
@@ -73,7 +74,7 @@ import org.tanukisoftware.wrapper.WrapperManager;
  *  server.stop();
  * </pre>
  *
- * @author Leif Mortenson <leif@tanukisoftware.com>
+ * @author Tanuki Software Development Team &lt;support@tanukisoftware.com&gt;
  */
 public class WrapperActionServer
     implements Runnable
@@ -129,7 +130,27 @@ public class WrapperActionServer
         m_port = port;
         m_bindAddr = bindAddress;
         
-        m_out = new WrapperPrintStream( System.out, "WrapperActionServer: " );
+        boolean streamSet = false;
+        if ( "true".equals( System.getProperty( "wrapper.use_sun_encoding" ) ) ) {
+            String sunStdoutEncoding = System.getProperty( "sun.stdout.encoding" );
+            if ( ( sunStdoutEncoding != null ) && !sunStdoutEncoding.equals( System.getProperty( "file.encoding" ) ) ) {
+                /* We need to create the stream using the same encoding as the one used for stdout, else this will lead to encoding issues. */
+                try
+                {
+                    m_out = new WrapperPrintStream( System.out, false, sunStdoutEncoding, "WrapperActionServer: " );
+                    streamSet = true;
+                }
+                catch ( UnsupportedEncodingException e )
+                {
+                    /* This should not happen because we always make sure the encoding exists before launching a JVM. */
+                    System.out.println( WrapperManager.getRes().getString( "Failed to set the encoding '{0}' when creating a WrapperPrintStream.\n Make sure the value of sun.stdout.encoding is correct.", sunStdoutEncoding ) );
+                }
+            }
+        }
+        if ( !streamSet )
+        {
+            m_out = new WrapperPrintStream( System.out, "WrapperActionServer: " );
+        }
     }
     
     /**
@@ -269,6 +290,7 @@ public class WrapperActionServer
     
     /**
      * Stops the runner thread, blocking until it has stopped.
+     * @throws Exception If the thread is unable to interrupt.
      */
     public void stop()
         throws Exception
@@ -331,6 +353,8 @@ public class WrapperActionServer
     /**
      * Unregisters an action with the given command.  If no action exists with
      *  the specified command, the method will quietly ignore the call.
+     * 
+     * @param command Command to be unregistered.
      */
     public void unregisterAction( byte command )
     {
